"""
Provide powerful, flexible access to tablular data stored in an Excel sheet.

The canonical format for an Excel table is row 1 having headers; tabular content being contiguous and complete
in rows 2-x; and no non-tabular data present in the sheet.

A lot of the functionality of this file is dedicated to auto-detecting the edges and column headers for tables
with non-canonical formatting.  Lots of heuristics. Not [yet] tested.

Uses the cheap, lightweight xlrd-like class as an access layer.
"""


from .xlrd_like import XL_CELL_EMPTY, XL_CELL_TEXT, XL_CELL_NUMBER, XlrdSheetLike
from xlrd.biffh import XL_CELL_ERROR

N_OPTS = 4
(MULTI, ROW_GAPS, COL_GAPS, MATRIX) = range(N_OPTS)

def _mk_xl_opts():
    return [None] * N_OPTS


def chunks(xcol):
    """
    Finds contiguous chunks of data, identified as non-empty cells, in an iterable of xlrd cells
    Generates a list of 2-tuples (starting row, chunk length)
    :param xcol: iterable of cells
    :return:
    """
    st = None
    for i, k in enumerate(xcol):
        if k.ctype == XL_CELL_EMPTY:
            if st is not None:  # falling edge
                yield st, i - st
                st = None
        else:
            if st is None:  # rising edge
                st = i
    if st is not None:
        yield st, len(xcol) - st


def share(xcol):
    """
    Returns the fraction of the input data that is non-empty, and the first non-empty cell
    :param xcol:
    :return:
    """
    first = None
    c = d = 0
    for k in xcol:
        c += 1
        if k.ctype == XL_CELL_EMPTY:
            continue
        d += 1
        if first is None:
            first = c - 1
    return d/c, first


def _longest(xcol):
    """
    returns row, len for longest len in the iterable
    :param xcol:
    :return:
    """
    try:
        return sorted(chunks(xcol), key=lambda _x: _x[1], reverse=True)[0]
    except IndexError:
        return 0, 0


def clean_value(cell):
    if cell.ctype == XL_CELL_TEXT:
        val = cell.value.strip()
    else:
        val = cell.value
    return val


class _EmptyRow(Exception):
    pass


class XlSheet(XlrdSheetLike):
    """
    This class handles access to a single SHEET_NAME---
    Fully defined, the SHEET_NAME has a data row (default 1), data column (default 0), and a set of column headers
    """
    def _next_row_thresh(self, start=0, thresh=0.7):
        while start < self._s.nrows:
            sh, first = share(self._s.row(start))
            if sh > thresh:
                return start, first
            start += 1
        return 0, None

    def _next_col_thresh(self, start=0, thresh=0.6):
        """
        Here we monkey a little bit to ignore long header blocks over short data
        :param start:
        :param thresh:
        :return:
        """
        apparent_firstrow, apparent_start = self._next_row_thresh()
        use_thresh = (self._s.nrows - apparent_firstrow) * thresh

        if apparent_start and (apparent_start > start):
            start = apparent_start

        while start < self._s.ncols:
            row, ck = _longest(self._s.col(start))
            if ck > use_thresh:
                return start, row
            start += 1
        return 0, 0

    def _discover(self, datarow, datacol):
        """
        This is the magic part.  We use the chunks method above to identify contiguous data chunks.
        We call 'datacol' the first auto-detected column that has a single chunk accounting for over half the
        number of rows in the spreadsheet.

        We will write this as a set of naive cases and then consolidate it as/if opportunities to do so appear
        :return:
        """
        # defaults
        dr = 1
        dc = 0
        hr = 1

        if self.multi:
            # for multi, if no row gaps, we assume long data blocks unless they are specified
            # headerrow is ignored in multi case
            if self._getopt(ROW_GAPS):
                print('Not handling row gaps!')
            else:
                dc_det, dr_det = self._next_col_thresh(thresh=0.5)
                if datacol is None:
                    dc = dc_det
                else:
                    dc = int(datacol)

                if datarow is None:
                    dr = dr_det
                else:
                    dr = int(datarow)
                    hr = dr - 1

            if dr == 0:
                # second attempt: look for full row as header; next full row as data row
                #hr, dc_det = self._next_row_thresh()
                ## if dc_det > dc:
                ##    dc = dc_det

                dr, check = self._next_row_thresh(start=hr + 1)
                if check != dc:
                    print('warning: blank leading entries in first detected data row')

        else:
            # default case-- should handle strict
            # define header_row as first row whose longest stretch exceeds 80% of the spreadsheet width
            hr, dc_det = self._next_row_thresh()
            if datacol is None:
                dc = dc_det
            else:
                dc = int(datacol)

            if datarow is not None:
                dr = int(datarow)
            else:
                if self._getopt(ROW_GAPS):
                    dr, check = self._next_row_thresh(start=hr + 1)
                    if check != dc:
                        print('warning: blank leading entries in first detected data row')
                else:
                    dr = hr + 1
        self.datarow = dr
        self.datacol = dc
        self.headerrow = hr

    def _setopt(self, opt, val):
        self._opts[opt] = bool(val)
        # reset internal lastrow
        self._lr_int = None
        # recompute headers
        self._compute_headers()

    def _getopt(self, opt):
        return self._opts[opt]

    def set_option(self, option, value):
        if not isinstance(option, int):
            option = {'mu': MULTI,
                      'ro': ROW_GAPS,
                      'co': COL_GAPS,
                      'ma': MATRIX}[str(option).lower()[:2]]
        self._setopt(option, value)

    def __init__(self, sheet, strict=False, datarow=None, datacol=None, #  headerrow=None,
                 multiheader=False,
                 row_gaps=False,
                 col_gaps=False):
        """

        :param sheet:
        :param strict: [False] if true, strict-tabular defaults are assumed: datarow=1, datacol=0, multi=False,
         unless overridden at the command line.
         if false, discovery is attempted for any non-specified params
        :param datarow:
        :param datacol:
        :param multiheader:
        """
        self._s = sheet
        self._r = None
        self._lr = None
        self._lr_int = None
        self._hr = None
        self._c = None

        # don't know what I'm doing with this
        self._opts = _mk_xl_opts()
        self._setopt(MULTI, multiheader)
        self._setopt(ROW_GAPS, row_gaps)
        self._setopt(COL_GAPS, col_gaps)
        self._cached_headers = []

        if strict:
            self.datarow = datarow or 1
            self.datacol = datacol or 0
            self.headerrow = self.datarow - 1
            self._lr_int = self.nrows
        else:
            self._discover(datarow, datacol)

    @property
    def is_null(self):
        return self._s.nrows == 0

    @property
    def name(self):
        return self._s.name

    @property
    def nrows(self):
        return self._s.nrows

    @property
    def ncols(self):
        return self._s.ncols

    @property
    def datarow(self):
        return self._r

    @datarow.setter
    def datarow(self, row):
        self._r = int(row)

    @property
    def headerrow(self):
        if self._hr is None:
            return self.datarow - 1
        return self._hr

    @headerrow.setter
    def headerrow(self, row):
        self._hr = int(row)
        self._compute_headers()

    @property
    def headers(self):
        return self._cached_headers

    @property
    def datacol(self):
        return self._c

    @datacol.setter
    def datacol(self, col):
        self._c = int(col)
        self._lr_int = None
        self._compute_headers()

    @property
    def lastrow(self):
        if self._lr is not None:
            return self._lr
        if self._lr_int is None:
            if self._getopt(ROW_GAPS):
                # if ROW_GAPS is true, lastrow is the last row with a nonempty entry in the data column
                self._lr_int = max(i for i, k in enumerate(self._s.col(self.datacol))
                                   if k.ctype != XL_CELL_EMPTY) + 1
            else:
                # if ROW_GAPS is false: lastrow is the last row before the first empty row after the first data row
                try:
                    self._lr_int = next(i for i, k in enumerate(self._s.col(self.datacol))
                                        if i > self.datarow and k.ctype == XL_CELL_EMPTY)
                except StopIteration:
                    self._lr_int = self._s.nrows
        return self._lr_int

    @lastrow.setter
    def lastrow(self, value):
        if value is None:
            self._lr = value
            self._lr_int = None
        else:
            self._lr = min([int(value), self._s.nrows])

    @property
    def multi(self):
        return self._getopt(MULTI)

    def _header(self, i, multi, start=None):
        """
        construct header i based on multi-row header specification and current headerrow
        :param i:
        :param multi:
        :param start:
        :return:
        """
        st_m = start or 0
        st_s = start or self.headerrow

        if multi:
            val = ' '.join(str(self._s.cell(k, i).value).strip() for k in range(st_m, self.datarow)).strip()
        else:
            val = clean_value(self._s.cell(st_s, i))
        return val

    def _compute_headers(self, multi=None, start=None):
        """
        Generate a list of headers from the current configuration (slow)
        access the protected variable _headers to use the cached list (fast)

        :param multi:
        :param start: if multi is false, start is the header row. if multi is true, start is the start of the header
        :return:
        """
        if self.datacol is None:
            return
        multi = multi or self.multi
        headers = []
        for i in range(self.datacol, self._s.ncols):
            headers.append(self._header(i, multi, start))
        self._cached_headers = headers

    def _read_row(self, rownum, _make_dict=None):
        _o = []
        _empty = True
        for i, k in enumerate(self._s.row(rownum)):
            if i < self.datacol:
                continue
            if k.ctype == XL_CELL_TEXT:
                _empty = False
                _o.append(k.value.strip())
            elif k.ctype == XL_CELL_ERROR:
                _o.append('Error:%d' % k.value)
            elif k.ctype == XL_CELL_EMPTY:
                _o.append(None)
            else:
                _empty = False
                _o.append(k.value)

        if _empty and self._getopt(ROW_GAPS):
            raise _EmptyRow

        if _make_dict is not None:
            return {k: v for k, v in zip(_make_dict, _o)}
        return _o

    def get_rows(self):
        for i, row in self.gen_rows():
            yield row

    def gen_rows(self, mask=None, rowdict=False):
        """
        Blank cells have value None
        :param mask:
        :param rowdict:
        :return: generates i, row tuples, but only for [non-blank] data rows
        """
        if rowdict:
            h = self.headers
        else:
            h = None
        for i in range(self.datarow, self.lastrow):
            in_mask = i - self.datarow
            if mask is not None:
                if not mask[in_mask]:
                    continue
            try:
                yield i, self._read_row(i, _make_dict=h)
            except _EmptyRow:
                continue

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._s.row(item + self.datarow)[self.datacol:self._s.ncols]
        elif isinstance(item, tuple):
            dat = self._s.row(item[0] + self.datarow)[self.datacol:self._s.ncols]
            return dat[self.find_column(item[1])].value

    def __call__(self, item):
        return self._header(int(item), self.multi)

    def find_column(self, column):
        try:
            return int(column)
        except ValueError:
            try:
                return self.headers.index(column)
            except ValueError:
                try:
                    return next(i for i, k in enumerate(self.headers) if k is not None and k.startswith(column))
                except StopIteration:
                    raise KeyError('Column %s not found' % column)

    def _find_column(self, column):
        return self.find_column(column) + self.datacol

    def row(self, row, rowdict=False):
        if rowdict:
            h = self.headers
        else:
            h = None
        return self._read_row(row + self.datarow, _make_dict=h)

    def row_dict(self, row):
        return self.row(row, rowdict=True)

    def col(self, column, mask=None):
        column = self._find_column(column)
        if mask is None:
            return self._s.col(column)[self.datarow:self.lastrow]
        else:
            dat = self._s.col(column)[self.datarow:self.lastrow]
            return [k for i, k in enumerate(dat) if mask[i]]

    def cell(self, row, col):
        return self._s.cell(row, col)

    def col_data(self, column, mask=None):
        return [clean_value(k) for k in self.col(column, mask=mask)]

    def total(self, column, mask=None):
        return sum(k.value for k in self.col(column, mask=mask) if k.ctype == XL_CELL_NUMBER)

    def unique(self, *columns, mask=None):
        if len(columns) == 1:
            try:
                return sorted(set(self.col_data(columns[0], mask=mask)))
            except TypeError:
                return set(self.col_data(columns[0], mask=mask))
        else:
            try:
                return sorted(set(zip(*(self.col_data(column, mask=mask) for column in columns))))  # that's some "pythonic" notation
            except TypeError:
                return set(zip(*(self.col_data(column, mask=mask) for column in columns)))

    def to_dataframe(self, mask=None, **kwargs):
        import pandas as pd
        return pd.DataFrame({k: self.col_data(i, mask=mask) for i, k in enumerate(self.headers)}, **kwargs)





import openpyxl
from .xlrd_like import XlrdSheetLike, XlrdCellLike, XlrdWorkbookLike


class OpenpyxlSheetLike(XlrdSheetLike):

    @property
    def xlsx(self):
        """
        Allow native access to openpyxl object
        :return:
        """
        return self._xlsx

    def __init__(self, xlsx_sheet):
        """
        We have to access the openpyxl sheet's internal dict of _cells in order to filter out Nones when computing
        the spreadsheet's range
        :param xlsx_sheet:
        """
        self._xlsx = xlsx_sheet
        max_row = max_col = 0
        for coord, cell in xlsx_sheet._cells.items():  # we have to do this because openpyxl is.. designed for purposes different from mine
            if cell.data_type == 'n':
                continue
            max_row = max([max_row, coord[0]])
            max_col = max([max_col, coord[1]])
        self._nrows = max_row
        self._ncols = max_col

    @property
    def name(self):
        return self._xlsx.title

    @property
    def ncols(self):
        return self._ncols

    @property
    def nrows(self):
        return self._nrows

    def row(self, row):
        """
        zero-indexed!
        :param row:
        :return:
        """
        row += 1
        if row > self._nrows:
            raise IndexError

        rows = list(self._xlsx.iter_rows(min_row=row, max_row=row, max_col=self._ncols))  # 2nd order list)
        return [XlrdCellLike(k.value) for k in rows[0]]

    def get_rows(self):
        rows = list(self._xlsx.iter_rows(min_row=1, max_row=self._nrows, max_col=self._ncols))
        for row in rows:
            yield [XlrdCellLike(k.value) for k in row]

    def col(self, col):
        """
        Zero-indexed!
        AND, per xlrd, negative index is counting from ncols
        :param col:
        :return:
        """
        if col < 0:
            col = self.ncols + col + 1
        else:
            col += 1
        if col > self._ncols:
            raise IndexError

        cols = list(self._xlsx.iter_cols(min_col=col, max_col=col, max_row=self._nrows))
        return [XlrdCellLike(k.value) for k in cols[0]]

    def cell(self, row, col):
        row += 1
        col += 1
        if row > self._nrows or col > self._ncols:
            raise IndexError

        cell = self._xlsx.cell(row, col)
        return XlrdCellLike(cell.value)


class OpenpyXlrdWorkbook(XlrdWorkbookLike):

    @classmethod
    def from_file(cls, file, **kwargs):
        return cls(openpyxl.load_workbook(file, **kwargs))

    @property
    def book(self):
        """
        Allow native access to book
        :return:
        """
        return self._book

    def __init__(self, xl_book):
        """
        :param xl_book: an initialized Openpyxl Workbook
        """
        self._book = xl_book
        self._names = {k: i for i, k in enumerate(self._book.sheetnames)}
        self._sheets = [OpenpyxlSheetLike(self._book[k]) for k in self._book.sheetnames]

    def sheet_names(self):
        return self._book.sheetnames

    def sheet_by_name(self, name):
        return self._sheets[self._names[name]]

    def sheet_by_index(self, index):
        return self._sheets[index]

    def sheets(self):
        return self._sheets

    @property
    def filename(self):
        return self._book.file



import os

try:
    # don't know how to test this in CI both with and without pandas (other than just test; pip install pandas; test again)
    import pandas as pd
except ImportError:
    from .pd_emulator import PandasEmulator as pd


from .xlrd_like import XlrdCellLike, XlrdSheetLike, XlrdWorkbookLike


def _make_cell(val):
    if pd.isna(val):
        return XlrdCellLike(None)
    return XlrdCellLike(val)


class CsvSheet(XlrdSheetLike):
    """
    Import a CSV and interact with it like an xlrd workbook containing only one sheet.

    Uses pandas to import the csv if pandas is available; otherwise, uses a minimal pandas DataFrame emulator.
    """
    def __init__(self, csvfile, **kwargs):
        name, ext = os.path.splitext(os.path.basename(csvfile))
        if ext.lower() != '.csv':
            print('Does not appear to be a csv: %s' % ext)
        self._name = name
        self._df = pd.read_csv(csvfile, **kwargs)

        self._headers = list(self._df.columns)

    def _find_column(self, column):
        """
        if number,
        :param column:
        :return:
        """
        if isinstance(column, int):
            return self._headers[column]
        elif column in self._headers:
            return column
        raise KeyError(column)

    @property
    def name(self):
        return self._name

    @property
    def nrows(self):
        return len(self._df) + 1

    @property
    def ncols(self):
        return len(self._headers)

    def row(self, row):
        if row == 0:
            return [_make_cell(k) for k in self._headers]
        else:
            return [_make_cell(c) for c in self._df.loc[row - 1]]

    def col(self, col):
        f = self._find_column(col)
        return [_make_cell(f)] + [_make_cell(c) for c in self._df[f]]

    def cell(self, row, col):
        return self.row(row)[col]

    def get_rows(self):
        for i in range(self.nrows):
            yield self.row(i)


class CsvWorkbook(XlrdWorkbookLike):
    def __init__(self, csvfile, **kwargs):
        self._csvfile = csvfile
        self._xl = CsvSheet(csvfile, **kwargs)

    @property
    def filename(self):
        return os.path.basename(self._csvfile)

    def sheet_names(self):
        return [self._xl.name]

    def sheet_by_index(self, index):
        if index == 0:
            return self._xl
        raise IndexError

    def sheet_by_name(self, name):
        if name == self._xl.name:
            return self._xl
        raise KeyError

    def sheets(self):
        return [self._xl]

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.sheet_by_index(item)
        return self.sheet_by_name(item)



"""
The purpose of these tools is to normalize access to tabular data in excel format.

The XlSheet class takes in a single sheet along with specifications for how to interpret the data-- key parameters
are the location of the data region (first row and first column of data).  Other parameters deal with specialty
aspects of interpreting header columns, and possibly data but for now that is ad hoc.

The XlSheet class provides some useful functions:
 - return headers (column names)
 - return record (row) as a list
 - return column as a list
 - return total of numeric columns
 -

For the moment, the
"""


import os

from .open_xl import open_xl
from .xlrd_like import XlrdWorkbookLike
from .xl_sheet import XlSheet


class XlReader(XlrdWorkbookLike):
    """
    How many times and in how many variants has this class been created?
    """
    def _get_sheet_index(self, sheet):
        if isinstance(sheet, int):
            return sheet
        try:
            inx = self._xl.sheet_names().index(sheet)
        except IndexError:
            try:
                inx = next(i for i, k in enumerate(self._xl.sheet_names()) if k.startswith(sheet))
            except StopIteration:
                raise KeyError('Sheet not found %s' % sheet)
        return inx

    def select_sheet(self, sheet):
        return self.__getitem__(sheet)

    def _check_xl_sheet(self, inx):
        if not isinstance(self._sheets[inx], XlSheet):
            self._sheets[inx] = XlSheet(self._xl.sheet_by_index(inx), **self._args)
        return self._sheets[inx]

    def __getitem__(self, item):
        inx = self._get_sheet_index(item)
        if inx is None:
            raise KeyError
        return self._check_xl_sheet(inx)

    def __init__(self, xlfile, formatting_info=False, **kwargs):
        """
        Open an Xl file for tabular data access
        :param xlfile: an XlrdWorkbookLike or a filename
        :param formatting_info: whether to open the spreadsheet with formatting (not implemented upstream for XLSX)
        :param kwargs: defaults to get passed to every XlSheet
        """
        self._args = kwargs
        if isinstance(xlfile, XlrdWorkbookLike):
            self._xl = xlfile
            self._fname = xlfile.filename
        else:
            self._xl = open_xl(xlfile, formatting_info=formatting_info)
            self._fname = os.path.abspath(xlfile)

        self._sheets = [None] * len(self._xl.sheet_names())

    @property
    def filepath(self):
        return self._fname

    @property
    def filename(self):
        return os.path.basename(self._fname)

    def __len__(self):
        return len(self._xl.sheets())

    def gen_rows(self, sheet=None):
        sh = self.__getitem__(sheet)
        return sh.gen_rows()

    @property
    def sheet_names(self):
        return self._xl.sheet_names()

    def sheet_by_name(self, name):
        return self.__getitem__(name)

    def sheet_by_index(self, index):
        return self.__getitem__(index)

    def sheets(self):
        return [self.__getitem__(k) for k in self.sheet_names]



"""
No need to introduce a dependency on pandas just to import csv files.  This is a minimal implementation of the pandas
dataframe that is produced from pd.read_csv().  It includes a list of NA values that pandas automatically detects.
"""

import csv

_NA_VALUES = [  # taken from pandas._libs.parsers.py
    b'',
    b'-NaN',
    b'#N/A N/A',
    b'-1.#IND',
    b'1.#QNAN',
    b'N/A',
    b'<NA>',
    b'null',
    b'NA',
    b'NaN',
    b'nan',
    b'NULL',
    b'-nan',
    b'-1.#QNAN',
    b'#NA',
    b'1.#IND',
    b'n/a',
    b'#N/A',
]


class PandasEmulator(object):
    na_values = [k.decode('utf8') for k in _NA_VALUES]

    @classmethod
    def read_csv(cls, csvfile, quoting=csv.QUOTE_MINIMAL, **kwargs):
        with open(csvfile, 'r') as fp:
            dr = csv.reader(fp, quoting=quoting, **kwargs)
            return cls(dr)

    def __init__(self, iter_row):
        self._header = list(next(iter_row))
        self._data = list(iter_row)

    @property
    def columns(self):
        return self._header

    def __len__(self):
        return len(self._data)

    @property
    def loc(self):
        return self._data

    def __getitem__(self, item):
        inx = self._header.index(item)
        for row in self._data:
            yield row[inx]

    @classmethod
    def isna(cls, value):
        return value is None or value in cls.na_values



from googleapiclient import discovery
from googleapiclient.http import HttpError
from oauth2client.service_account import ServiceAccountCredentials
from .xlrd_like import XlrdCellLike, XlrdSheetLike, XlrdWriteWorkbook
from .util import colnum_to_col

import time


class GoogleSheetError(Exception):
    pass


class GSheetCell(XlrdCellLike):
    def __init__(self, str_value):
        if len(str_value) == 0:
            value = None
        else:
            # detect type: either blank, number, or string
            try:
                value = float(str_value)
            except (TypeError, ValueError):
                value = str_value
        super(GSheetCell, self).__init__(value)


class GSheetEmulator(XlrdSheetLike):
    def __init__(self, value_data):
        """

        :param value_data: must follow the google sheets v4 API .spreadsheets().values().get(... range=sheetname)
        """
        if len(value_data['values']) == 0:
            _nr = _nc = 0
            data = [[]]
        else:
            if value_data['majorDimension'] == 'ROWS':
                data = value_data['values']
            elif value_data['majorDimension'] == 'COLUMNS':
                raise NotImplementedError
            else:
                raise GoogleSheetError('Cannot interpret value_data majorDimension')
            _nr = len(data)
            _nc = max([len(k) for k in data])

        self._range = value_data['range']

        self._nr = int(_nr)
        self._nc = int(_nc)
        self._data = data

    @property
    def name(self):
        return self._range.split('!')[0].strip('\'\"')

    @property
    def range(self):
        return self._range.split('!')[1]

    @property
    def nrows(self):
        return self._nr

    @property
    def ncols(self):
        return self._nc

    def row(self, row):
        return list(GSheetCell(k) for k in self._data[row])

    def get_rows(self):
        for i in range(self.nrows):
            yield self.row(i)

    def col(self, col):
        cd = []
        for row in range(self.nrows):
            try:
                cd.append(GSheetCell(self._data[row][col]))
            except IndexError:
                cd.append(GSheetCell(''))
        return cd

    def cell(self, row, col):
        return GSheetCell(self._data[row][col])


class GoogleSheetReader(XlrdWriteWorkbook):
    """
    Creates an xlrd-like google sheet reader with the following properties:

    sheet_by_name(sheet): returns a 'sheet-like' object

    'sheet-like' object has the following minimal API:
    nrows: (property) returns int number of rows
    ncols: (property) returns int number of columns
    row(n): returns a list of 'cell-like' objects in row n (0-indexed)

    'cell-like' object has the following minimal API:
    ctype: (property) integer 0=empty, 1=text, 2=number, 3=date, 4=bool, 5=error, 6=debug
    value: (property) value of the cell

    Also confers writing abilities (much easier to write by API than to write an old-fashioned file)

    """
    def __init__(self, credentials, sheet_id):
        """
        Creates an Xlrd-like object that also has create-sheet and write-to-sheet capabilities.

        For instructions on obtaining a JWT credential file, please visit:
         https://docs.gspread.org/en/latest/oauth2.html#service-account

        the sheet_id is the long alphanumeric string that appears in the URL, e.g.:
        https://docs.google.com/spreadsheets/d/{sheet_id}/edit#...

        You must grant your service account authority to access the sheet using the "Share" button.

        :param credentials: either a path to a credential file, or a credential dict (as derived from a file)
        :param sheet_id:
        """
        if isinstance(credentials, dict):
            cred = ServiceAccountCredentials.from_json_keyfile_dict(credentials,
                                                                    scopes=['https://spreadsheets.google.com/feeds'])
        else:
            cred = ServiceAccountCredentials.from_json_keyfile_name(credentials,
                                                                    scopes=['https://spreadsheets.google.com/feeds'])
        self._res = discovery.build('sheets', 'v4', credentials=cred)

        self._sheet_id = sheet_id

        self._sheetnames = self.sheet_names()

    @property
    def filename(self):
        return self._sheet_id

    def sheet_names(self):
        req = self._res.spreadsheets().get(spreadsheetId=self._sheet_id)
        d = req.execute()
        return [k['properties']['title'] for k in d['sheets']]

    def sheet_by_name(self, sheetname):
        """
        Runs a new request every time- no caching
        :param sheetname:
        :return:
        """
        quoted_sheetname = "'%s'" % sheetname  # without quotes it may be interpreted as a named range
        req = self._res.spreadsheets().values().get(spreadsheetId=self._sheet_id, range=quoted_sheetname)
        try:
            d = req.execute()
        except HttpError:
            raise KeyError('Unable to open sheet %s' % sheetname)
        return GSheetEmulator(d)

    def sheet_by_index(self, index):
        return self.sheet_by_name(self._sheetnames[index])

    def sheets(self):
        """
        No sheet caching!
        :return:
        """
        return [self.sheet_by_name(name) for name in self._sheetnames]

    def create_sheet(self, name, **kwargs):
        kwargs['title'] = name

        body = {'requests': [
            {'addSheet':
                 {'properties': kwargs}
             }

        ]}
        req = self._res.spreadsheets().batchUpdate(spreadsheetId=self._sheet_id,
                                                   body=body)
        ret = req.execute()
        self._sheetnames = self.sheet_names()
        return ret

    def write_to_sheet(self, sheet, range, data, **kwargs):
        """
        The data must be a 2d array that matches the size of the range argument
        :param sheet:
        :param range:
        :param data:
        :param kwargs: added to request body
        :return:
        """
        r = '%s!%s' % (sheet, range)
        kwargs['values'] = data
        req = self._res.spreadsheets().values().update(spreadsheetId=self._sheet_id, range=r,
                                                       body=kwargs, valueInputOption='RAW')
        result = req.execute()
        time.sleep(1)  # standard quota is only 60 requests per minute per user (300 per minute per project)
        # use write_rectangle_by_rows and [nonimpl] write_rectangle_by_columns
        return result

    def write_cell(self, sheet, row, col, value, **kwargs):
        """

        :param sheet:
        :param row: 0-indexed
        :param col: 0-indexed number or alphabetical column name e.g. 'A'
        :param value:
        :param kwargs: added to request body
        :return:
        """
        col = colnum_to_col(col)
        data = [[value]]
        rn = '%s%d:%s%d' % (col, row + 1, col, row + 1)
        return self.write_to_sheet(sheet, rn, data, **kwargs)

    def write_column(self, sheet, col, values, start_row=0, **kwargs):
        """
        Write sequential data into a column of the google sheet.
        :param sheet:
        :param col: either a column string (e.g. 'AA') or a 0-indexed column number (e.g. 0 = 'A', 1 = 'B', ...)
        :param values:
        :param start_row: 0-indexed row to begin (note: google-sheets are 1-indexed so start_row=0 corresponds to row 1)
        :param kwargs:
        :return:
        """
        col = colnum_to_col(col)
        data = [[k] for k in values]
        n = len(data)
        rn = '%s%d:%s%d' % (col, start_row+1, col, start_row + n)
        self.write_to_sheet(sheet, rn, data, **kwargs)

    def write_row(self, sheet, row, values, start_col=0, **kwargs):
        """
        Write sequential data into a row of the google sheet.  (Note: google-sheets are 1-indexed so row = 0
        corresponds to the spreadsheet's native row 1)
        :param sheet:
        :param row:
        :param values:
        :param start_col: 0-indexed Column to begin (default 0 / 'A')
        :param kwargs:
        :return:
        """
        row += 1
        data = [[k for k in values]]
        n = len(data[0])
        rn = '%s%d:%s%d' % (colnum_to_col(start_col), row, colnum_to_col(start_col + n - 1), row)
        self.write_to_sheet(sheet, rn, data, **kwargs)

    def write_rectangle_by_rows(self, sheet, row_gen, start_row=0, start_col=0, **kwargs):
        """
        Write data to a rectangular area, starting at start_row and start_col.  The size of the rectangle
        is determined by the longest row {{short rows are padded with Nones, which are ignored by gsheet API}}

        This is vital to avoiding unbearably slow execution, due to google's rate limit of 60 queries/minute/user

        :param sheet:
        :param row_gen: a generator that produces iterables of values for each row, beginning with start_col
        :param start_row: 0-indexed start row
        :param start_col: 0-indexed start column
        :param kwargs:
        :return:
        """
        end_row = start_row
        start_row += 1
        data = []
        n = 0
        for row in row_gen:
            nextdata = [value for value in row]
            n = max([n, len(nextdata)])
            data.append(nextdata)
            end_row += 1

        if end_row < start_row:  # no rows generated
            print('write_rectangle: no data provided')
            return

        for row in data:
            while len(row) < n:
                row.append(None)

        rn = '%s%d:%s%d' % (colnum_to_col(start_col), start_row, colnum_to_col(start_col + n - 1), end_row)
        self.write_to_sheet(sheet, rn, data, **kwargs)

    def clear_region(self, sheet, start_row=0, start_col=0, end_row=None, end_col=None, **kwargs):
        """
        Clear the region using the gsheet API.  Note: input args are 0-indexed, noting that API is 1-indexed.
        Default is to clear the entire sheet.
        :param sheet: must exist
        :param start_row: 0-indexed. defaults to first row
        :param start_col: 0-indexed. defaults to first column
        :param end_row: 0-indexed. defaults to last row
        :param end_col: 0-indexed. defaults to last column
        :param kwargs: passed as request body
        :return:
        """
        s = self.sheet_by_name(sheet)
        if end_row is None or end_row > (s.nrows - 1):
            end_row = s.nrows
        else:
            end_row += 1
        if end_col is None or end_col > (s.ncols - 1):
            end_col = s.ncols
        else:
            end_col += 1
        start_row = max([start_row + 1, 1])
        start_col = max([start_col + 1, 1])

        rn = '%s!R%dC%d:R%dC%d' % (sheet, start_row, start_col, end_row, end_col)
        req = self._res.spreadsheets().values().clear(spreadsheetId=self._sheet_id, range=rn, body=kwargs)
        req.execute()

    def write_dataframe(self, sheetname, df, clear_sheet=True, write_header=True, header_levels=None,
                        fillna='NA', write_index=True):
        """

        :param self: a GoogleSheetReader
        :param sheetname: sheet to write to or create
        :param df: a pandas dataframe
        :param clear_sheet: [True]
        :param write_header: [True] whether to write header (False: leave it standing)
        :param header_levels: number of header levels to write. Must be <= nlevels
        :param fillna:
        :param write_index:
        :return:
        """

        ncol = len(df.columns)
        if not write_index:
            ncol -= 1
        if header_levels is None or header_levels > df.columns.nlevels:
            header_levels = df.columns.nlevels

        if sheetname in self.sheet_names():
            # start by clearing the sheet- with or without headers
            if clear_sheet:
                if write_header:
                    self.clear_region(sheetname)
                else:
                    self.clear_region(sheetname, start_row=header_levels)
            else:
                if write_header:
                    self.clear_region(sheetname, end_col=ncol, end_row=header_levels - 1)
        else:
            self.create_sheet(sheetname)

        #then populate
        def _row_gen(_df):
            for _i, row in _df.fillna(fillna).iterrows():
                if write_index:
                    yield [_i] + list(row.values)
                else:
                    yield list(row.values)

        if write_header:
            for i in range(header_levels):
                if write_index:
                    h = [''] + list(df.columns.get_level_values(i))
                else:
                    h = list(df.columns.get_level_values(i))
                self.write_row(sheetname, i, h)
        self.write_rectangle_by_rows(sheetname, _row_gen(df), start_row=header_levels)



"""
This module presents an xlrd-like interface to an openpyxl (e.g. excel 2007+) spreadsheet.  The supported xlrd
API is as demanded by existing MFA code, and is as follows:

XlrdLike: workbook equivalent
 .sheet_names() - return list of sheet names
 .sheet_by_name() - return an XlSheetLike given by name
 .sheet_by_index() - return an XlSheetLike given by index
 .sheets() - return a list of XlSheetLikes-- requires initializing every sheet

XlSheetLike: worksheet equivalent
 .name - the name of the sheet according to the workbook
 .nrows - int number of rows, 0-indexed
 .ncols - int number of columns, 0-indexed
 .row(n) - return a list of XlCellLike corresponding to the nth (0-indexed) row, or IndexError
 .col(k) - return a list of XlCellLike corresponding to the kth (0-indexed) column, or IndexError
 .cell(n,k) - return the nth row, kth cell, or IndexError
 .get_rows() - row iterator

XlCellLike: cell equivalent
 .ctype - int, as indicated below
 .value - native value
"""
import abc
from datetime import datetime

import openpyxl
from xlrd.biffh import (
    XL_CELL_EMPTY,    # 0
    XL_CELL_TEXT,     # 1
    XL_CELL_NUMBER,   # 2
    XL_CELL_DATE,     # 3
    XL_CELL_BOOLEAN,  # 4
    # XL_CELL_ERROR, # 5
    # XL_CELL_BLANK, # 6 - for use in debugging, gathering stats, etc
)


class XlrdCellLike(object):
    """
    Subclass this to change how cells are interpreted as value + type
    """
    def __init__(self, cell):
        self._cell = cell

    @property
    def value(self):
        if self.ctype == XL_CELL_TEXT:
            return str(self._cell)
        return self._cell

    @property
    def ctype(self):
        if self._cell is None:
            return XL_CELL_EMPTY
        elif isinstance(self._cell, openpyxl.compat.NUMERIC_TYPES):
            # TODO: figure out how to detect excel-style dates
            return XL_CELL_NUMBER
        elif isinstance(self._cell, bool):
            return XL_CELL_BOOLEAN
        elif isinstance(self._cell, datetime):
            return XL_CELL_DATE
        else:
            return XL_CELL_TEXT


class XlrdSheetLike(object):
    @property
    def name(self):
        raise NotImplementedError

    @property
    def nrows(self):
        raise NotImplementedError

    @property
    def ncols(self):
        raise NotImplementedError

    def row(self, row):
        raise NotImplementedError

    def col(self, col):
        raise NotImplementedError

    def cell(self, row, col):
        raise NotImplementedError

    def get_rows(self):
        raise NotImplementedError

    def row_dict(self, row):
        """
        Creates a dictionary of the nth row using the 0th row as keynames
        TODO: need to assign this capability to native xlrd sheets, which don't have it
        :param row:
        :return:
        """
        headers = [k.value for k in self.row(0)]
        return {headers[i]: k.value for i, k in enumerate(self.row(row)[:len(headers)])}


class XlrdWorkbookLike(abc.ABC):
    """
     .sheet_names() - return list of sheet names
     .sheet_by_name() - return an XlSheetLike given by name
     .sheet_by_index() - return an XlSheetLike given by index
     .sheets() - return a list of XlSheetLikes-- requires initializing every sheet
    """
    def sheet_names(self):
        raise NotImplementedError

    def sheet_by_name(self, name):
        raise NotImplementedError

    def sheet_by_index(self, index):
        raise NotImplementedError

    def sheets(self):
        raise NotImplementedError

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.sheet_by_index(item)
        else:
            return self.sheet_by_name(item)

    @property
    def filename(self):
        raise NotImplementedError


class XlrdWriteWorkbook(XlrdWorkbookLike):
    """
     .create_sheet()
     .write_cell()
     .write_row()
     .write_column()
     .write_rectangle_by_rows()
     .clear_region()
    """
    def create_sheet(self, sheetname, **kwargs):
        raise NotImplementedError

    def write_cell(self, sheet, row, col, value, **kwargs):
        raise NotImplementedError

    def write_row(self, sheet, row, values, start_col, **kwargs):
        raise NotImplementedError

    def write_col(self, sheet, col, values, start_row, **kwargs):
        raise NotImplementedError

    def write_rectangle_by_rows(self, sheet, row_gen, start_row=0, start_col=0, **kwargs):
        raise NotImplementedError

    def clear_region(self, sheet, start_row=0, start_col=0, end_row=None, end_col=None, **kwargs):
        raise NotImplementedError



def col_to_colnum(col):
    """
    Convert an excel column name to a 0-indexed number 'A' = 0, 'B' = 1, ... 'AA' = 26, ... 'AAZ' = 727, ...
    :param col:
    :return:
    """
    if not isinstance(col, str):
        return int(col)
    cols = list(col.upper())
    num = 0
    while len(cols) > 0:
        num *= 26
        c = cols.pop(0)
        num += (ord(c) - ord('A') + 1)
    return num - 1


def colnum_to_col(num):
    """
    Convert a 0-indexed numeric index into an alphabetical column label. 0 = 'A', 1 = 'B'... 26 = 'AA', 27 = 'AB', ...
    :param num:
    :return:
    """
    if isinstance(num, str):
        return num
    col = ''
    num = int(num)
    while 1:
        rad = num % 26
        col = chr(ord('A') + rad) + col
        num //= 26
        num -= 1
        if num < 0:
            return col



from setuptools import setup, find_packages

requires = [
    "xlrd>=2.0.0",
    "openpyxl>=3.0.7"
]

"""
Version History:

0.1.0  - 8 April 2022 - initial public release

0.1.1 - 1 August 2022 - add CSV reader (with pandas emulation!)
        change package name to xlstools
        
0.1.2 - 19 February 2023 - finish csv; add write class

0.1.3 - 21 February 2023 - dumb abc error. we should think about writing some tests :P
"""


VERSION = '0.1.3'

setup(
    name="xlstools",
    version=VERSION,
    author="Brandon Kuczenski",
    author_email="brandon@scope3consulting.com",
    install_requires=requires,
    extras_require={
        'gsheet': ["google-api-python-client>=2.2.0", "oauth2client>=4.1.3"]
    },
    url="https://github.com/scope3/xls-tools",
    summary="Tricky tricks with XLS",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=find_packages()
)



import xlrd
from .openpyxlrd import OpenpyXlrdWorkbook
from .csv_reader import CsvWorkbook


def open_xl(path, formatting_info=False, data_only=True, **kwargs):
    """
    Reads XLS, XLSX, or CSV files into an object with a consistent, minimal read-only interface based on xlrd
    :param path:
    :param formatting_info:
    :param data_only:
    :param kwargs:
    :return:
    """
    if path.lower().endswith('xls'):
        return xlrd.open_workbook(path, formatting_info=formatting_info)
    elif path.lower().endswith('csv'):
        return CsvWorkbook(path, **kwargs)
    else:
        '''
        try:
        except:
        '''
        return OpenpyXlrdWorkbook.from_file(path, data_only=data_only, **kwargs)



"""
TODO: import CSV
"""

import os
import xlrd
import re

from datetime import datetime

try:
    from .google_sheet_reader import GoogleSheetReader
except ImportError:
    print('Unable to import GoogleSheetReader - try python setup.py install [gsheet]')
    GoogleSheetReader = False
from .xl_reader import XlReader
from .xl_sheet import XlSheet
from .openpyxlrd import OpenpyXlrdWorkbook
from .open_xl import open_xl
from .util import colnum_to_col, col_to_colnum
# from .exchanges_from_spreadsheet import exchanges_from_spreadsheet


def xl_date(cell_or_value, mode=0, short=True):
    """
    This uses an xlrd utility function to convert excel integer dates to date tuples.
    :param cell_or_value:
    :param mode:
    :param short:
    :return:
    """
    if isinstance(cell_or_value, xlrd.sheet.Cell):
        val = cell_or_value.value
    else:
        val = cell_or_value
    if isinstance(val, datetime):
        tup = val.timetuple()
    else:
        tup = xlrd.xldate_as_tuple(val, mode)
    if short:
        return tup[:3]
    else:
        return tup


def xls_files(_path):
    """
    Generates xls / xlsx files in a directory
    :param _path:
    :return:
    """
    for k in os.listdir(_path):
        if bool(re.match('^[^\.].+\.xlsx?$', k, flags=re.I)):
            yield os.path.join(_path, k)
